@@ -1,16 +1,18 @@
 package Gearman::Driver;
 
 use Moose;
+use Moose::Util qw(apply_all_roles);
+use Class::MOP;
 use Carp qw(croak);
 use Gearman::Driver::Observer;
+use Gearman::Driver::Console;
 use Gearman::Driver::Job;
 use Log::Log4perl qw(:easy);
-use Module::Find;
 use MooseX::Types::Path::Class;
 use POE;
-with qw(MooseX::Log::Log4perl MooseX::Getopt);
+with qw(MooseX::Log::Log4perl MooseX::Getopt Gearman::Driver::Loader);
 
-our $VERSION = '0.01000_02';
+our $VERSION = '0.01025_01';
 
 =head1 NAME
 
@@ -20,24 +22,25 @@ Gearman::Driver - Manages Gearman workers
 
     package My::Workers::One;
 
+    # Yes, you need to do it exactly this way
     use base qw(Gearman::Driver::Worker);
     use Moose;
 
     # this method will be registered with gearmand as 'My::Workers::One::scale_image'
     sub scale_image : Job {
-        my ( $self, $job ) = @_;
+        my ( $self, $job, $workload ) = @_;
         # do something
     }
 
     # this method will be registered with gearmand as 'My::Workers::One::do_something_else'
-    sub do_something_else : Job : MinChilds(2) : MaxChilds(15) {
-        my ( $self, $job ) = @_;
+    sub do_something_else : Job : MinProcesses(2) : MaxProcesses(15) {
+        my ( $self, $job, $workload ) = @_;
         # do something
     }
 
     # this method wont be registered with gearmand at all
     sub do_something_internal {
-        my ( $self, $job ) = @_;
+        my ( $self, $job, $workload ) = @_;
         # do something
     }
 
@@ -50,7 +53,7 @@ Gearman::Driver - Manages Gearman workers
 
     # this method will be registered with gearmand as 'My::Workers::Two::scale_image'
     sub scale_image : Job {
-        my ( $self, $job ) = @_;
+        my ( $self, $job, $workload ) = @_;
         # do something
     }
 
@@ -95,13 +98,13 @@ See also: L<namespaces|/namespaces>. If you do not set
 L<server|/server> (gearmand) attribute the default will be used:
 C<localhost:4730>
 
-Each module found in your namespace will be loaded and introspected,
+Each module found in your namespaces will be loaded and introspected,
 looking for methods having the 'Job' attribute set:
 
     package My::Workers::ONE;
 
     sub scale_image : Job {
-        my ( $self, $job ) = @_;
+        my ( $self, $job, $workload ) = @_;
         # do something
     }
 
@@ -126,7 +129,7 @@ a custom prefix:
     sub prefix { 'foo_bar_' }
 
     sub scale_image : Job {
-        my ( $self, $job ) = @_;
+        my ( $self, $job, $workload ) = @_;
         # do something
     }
 
@@ -136,33 +139,7 @@ See also: L<prefix|Gearman::Driver::Worker/prefix>
 
 =head1 ATTRIBUTES
 
-=head2 namespaces
-
-Will be passed to L<Module::Find> C<useall> method to load worker
-modules. Each one of those modules has to be inherited from
-L<Gearman::Driver::Worker> or a subclass of it. It's also possible
-to use the full package name to load a single module/file. There is
-also a method L<get_namespaces|Gearman::Driver/get_namespaces> which
-returns a sorted list of all namespaces.
-
-=over 4
-
-=item * isa: C<ArrayRef>
-
-=item * required: C<True>
-
-=back
-
-=cut
-
-has 'namespaces' => (
-    documentation => 'Example: --namespaces My::Workers --namespaces My::OtherWorkers',
-    handles       => { get_namespaces => 'sort' },
-    is            => 'rw',
-    isa           => 'ArrayRef[Str]',
-    required      => 1,
-    traits        => [qw(Array)],
-);
+See also L<Gearman::Driver::Loader/ATTRIBUTES>.
 
 =head2 server
 
@@ -189,12 +166,38 @@ has 'server' => (
     required      => 1,
 );
 
+=head2 console_port
+
+Gearman::Driver has a telnet management console, see also:
+
+L<Gearman::Driver::Console>
+
+=over 4
+
+=item * default: C<47300>
+
+=item * isa: C<Int>
+
+=back
+
+Set this to C<0> to disable management console at all.
+
+=cut
+
+has 'console_port' => (
+    default       => 47300,
+    documentation => 'Port of management console (default: 47300)',
+    is            => 'rw',
+    isa           => 'Int',
+    required      => 1,
+);
+
 =head2 interval
 
 Each n seconds L<Net::Telnet::Gearman> is used in
 L<Gearman::Driver::Observer> to check status of free/running/busy
 workers on gearmand. This is used to fork more workers depending
-on the queue size and the MinChilds/MaxChilds
+on the queue size and the MinProcesses/MaxProcesses
 L<attribute|Gearman::Driver::Worker/METHODATTRIBUTES> of the
 job method. See also: L<Gearman::Driver::Worker>
 
@@ -216,6 +219,38 @@ has 'interval' => (
     required      => 1,
 );
 
+=head2 max_idle_time
+
+Whenever L<Gearman::Driver::Observer> notices that there are more
+processes running than actually necessary (depending on min_processes
+and max_processes setting) it will kill them. By default this happens
+immediately. If you change this value to C<300>, a process which is
+not necessary is killed after 300 seconds.
+
+Please remember that this also depends on what value you set
+L</interval> to. The max_idle_time is only checked each n seconds
+where n is L</interval>. Besides that it makes only sense when you
+have workers where L<Gearman::Driver::Worker/MinProcesses> is set to
+C<0>.
+
+=over 4
+
+=item * default: C<0>
+
+=item * isa: C<Int>
+
+=back
+
+=cut
+
+has 'max_idle_time' => (
+    default       => '0',
+    documentation => 'How many seconds a worker may be idle before its killed',
+    is            => 'rw',
+    isa           => 'Int',
+    required      => 1,
+);
+
 =head2 logfile
 
 Path to logfile.
@@ -234,7 +269,7 @@ has 'logfile' => (
     coerce        => 1,
     default       => 'gearman_driver.log',
     documentation => 'Path to logfile (default: gearman_driver.log)',
-    is            => 'ro',
+    is            => 'rw',
     isa           => 'Path::Class::File',
 );
 
@@ -246,16 +281,16 @@ See also L<Log::Log4perl>.
 
 =item * isa: C<Str>
 
-=item * default: C<[%d] %m%n>
+=item * default: C<[%d] %p %m%n>
 
 =back
 
 =cut
 
 has 'loglayout' => (
-    default       => '[%d] %m%n',
-    documentation => 'Log message layout (default: [%d] %m%n)',
-    is            => 'ro',
+    default       => '[%d] %p %m%n',
+    documentation => 'Log message layout (default: [%d] %p %m%n)',
+    is            => 'rw',
     isa           => 'Str',
 );
 
@@ -276,7 +311,7 @@ See also L<Log::Log4perl>.
 has 'loglevel' => (
     default       => 'INFO',
     documentation => 'Log level (default: INFO)',
-    is            => 'ro',
+    is            => 'rw',
     isa           => 'Str',
 );
 
@@ -306,7 +341,7 @@ C<$status> might look like:
     $VAR1 = {
         'busy'    => 0,
         'free'    => 0,
-        'name'    => 'GDExamples::Sleeper::unknown_job',
+        'name'    => 'GDExamples::Convert::unknown_job',
         'queue'   => 6,
         'running' => 0
     };
@@ -326,35 +361,6 @@ has 'unknown_job_callback' => (
 
 This might be interesting for subclassing L<Gearman::Driver>.
 
-=head2 modules
-
-Every worker module loaded by L<Module::Find> will be added to this
-list. There are also two methods:
-L<get_modules|Gearman::Driver/get_modules> and
-L<has_modules|Gearman::Driver/has_modules>.
-
-=over 4
-
-=item * isa: C<ArrayRef>
-
-=item * readonly: C<True>
-
-=back
-
-=cut
-
-has 'modules' => (
-    default => sub { [] },
-    handles => {
-        _add_module => 'push',
-        get_modules => 'sort',
-        has_modules => 'count',
-    },
-    is     => 'ro',
-    isa    => 'ArrayRef[Str]',
-    traits => [qw(Array NoGetopt)],
-);
-
 =head2 jobs
 
 Stores all L<Gearman::Driver::Job> instances. The key is the name
@@ -385,7 +391,6 @@ has 'jobs' => (
     handles => {
         _set_job => 'set',
         get_job  => 'get',
-        get_jobs => 'values',
         has_job  => 'defined',
     },
     is     => 'ro',
@@ -413,12 +418,213 @@ has 'observer' => (
     traits => [qw(NoGetopt)],
 );
 
-has '+logger' => ( traits => [qw(NoGetopt)] );
+=head2 console
+
+Instance of L<Gearman::Driver::Console>.
+
+=over 4
+
+=item * isa: C<Gearman::Driver::Console>
+
+=item * readonly: C<True>
+
+=back
+
+=cut
+
+has 'console' => (
+    is     => 'ro',
+    isa    => 'Gearman::Driver::Console',
+    traits => [qw(NoGetopt)],
+);
+
+=head2 extended_status
+
+Enables/disables extended status information in
+L<management console|Gearman::Driver::Console::Basic/status> like
+lastrun, lasterror and lasterror_msg.
+
+=over 4
+
+=item * isa: C<Bool>
+
+=item * default: C<1>
+
+=back
+
+=cut
+
+has 'extended_status' => (
+    default       => 1,
+    documentation => 'Show extended status infos in management console',
+    is            => 'rw',
+    isa           => 'Bool',
+);
+
+has 'session' => (
+    is     => 'ro',
+    isa    => 'POE::Session',
+    traits => [qw(NoGetopt)],
+);
+
+# child communication socket
+has 'cc_socket' => (
+    default => "/tmp/gearman_driver-$$.sock",
+    is      => 'ro',
+    isa     => 'Str',
+    traits  => [qw(NoGetopt)],
+);
+
+has 'pid' => (
+    default => $$,
+    is      => 'ro',
+    isa     => 'Int',
+);
+
+has '+logger'  => ( traits => [qw(NoGetopt)] );
+has '+wanted'  => ( traits => [qw(NoGetopt)] );
+has '+modules' => ( traits => [qw(NoGetopt)] );
 
 =head1 METHODS
 
-Every method except L<run|/run> might only be interesting for
-people subclassing L<Gearman::Driver>.
+=head2 add_job
+
+There's one mandatory param (hashref) with following keys:
+
+=over 4
+
+=item * decode (optionally)
+
+Name of a decoder method in your worker object.
+
+=item * encode (optionally)
+
+Name of a encoder method in your worker object.
+
+=item * method (mandatory)
+
+Reference to a CodeRef which will get invoked.
+
+=item * min_processes (mandatory)
+
+Minimum number of processes that should be forked.
+
+=item * max_processes (mandatory)
+
+Maximum number of processes that may be forked.
+
+=item * name (mandatory)
+
+Job name/alias that method should be registered with Gearman.
+
+=item * object (mandatory)
+
+Object that should be passed as first parameter to the job method.
+
+=back
+
+Basically you never really need this method if you use
+L</namespaces>. But L</namespaces> depend on method attributes which
+some people do hate. In this case, feel free to setup your C<$driver>
+this way:
+
+    package My::Workers::One;
+
+    use Moose;
+    use JSON::XS;
+    extends 'Gearman::Driver::Worker::Base';
+
+    sub scale_image {
+        my ( $self, $job, $workload ) = @_;
+        # do something
+    }
+
+    # this method will be registered with gearmand as 'My::Workers::One::do_something_else'
+    sub do_something_else {
+        my ( $self, $job, $workload ) = @_;
+        # do something
+    }
+
+    sub encode_json {
+        my ( $self, $result ) = @_;
+        return JSON::XS::encode_json($result);
+    }
+
+    sub decode_json {
+        my ( $self, $workload ) = @_;
+        return JSON::XS::decode_json($workload);
+    }
+
+    1;
+
+    package main;
+
+    use Gearman::Driver;
+    use My::Workers::One;
+
+    my $driver = Gearman::Driver->new(
+        server   => 'localhost:4730,otherhost:4731',
+        interval => 60,
+    );
+
+    my $worker = My::Workers::One->new();
+
+    foreach my $method (qw(scale_image do_something_else)) {
+        $driver->add_job(
+            decode        => 'decode_json',
+            encode        => 'encode_json',
+            max_processes => 5,
+            method        => $worker->meta->find_method_by_name($method)->body,
+            min_processes => 1,
+            name          => $method,
+            object        => $worker,
+        );
+    }
+
+    $driver->run;
+
+
+=cut
+
+sub add_job {
+    my ( $self, $params ) = @_;
+
+    $params->{max_processes} = delete $params->{max_childs} if defined $params->{max_childs};
+    $params->{min_processes} = delete $params->{min_childs} if defined $params->{min_childs};
+
+    my $job = Gearman::Driver::Job->new(
+        driver        => $self,
+        decode        => $params->{decode} || '',
+        encode        => $params->{encode} || '',
+        max_processes => $params->{max_processes},
+        method        => $params->{method},
+        min_processes => $params->{min_processes},
+        name          => $params->{name},
+        server        => $self->server,
+        worker        => $params->{object},
+    );
+
+    $self->_set_job( $params->{name} => $job );
+
+    $self->log->debug( sprintf "Added new job: %s (processes: %d)", $params->{name}, $params->{min_processes} );
+
+    return 1;
+}
+
+=head2 get_jobs
+
+Returns all L<Gearman::Driver::Job> objects ordered by jobname.
+
+=cut
+
+sub get_jobs {
+    my ($self) = @_;
+    my @result = ();
+    foreach my $name ( sort keys %{ $self->jobs } ) {
+        push @result, $self->get_job($name);
+    }
+    return @result;
+}
 
 =head2 run
 
@@ -427,20 +633,33 @@ This must be called after the L<Gearman::Driver> object is instantiated.
 =cut
 
 sub run {
+    my ($self) = @_;
+    push @INC, @{ $self->lib };
+    $self->load_namespaces;
+    $self->_start_observer;
+    $self->_start_console;
+    $self->_start_session;
     POE::Kernel->run();
 }
 
-=head2 get_namespaces
+=head2 shutdown
 
-Returns a sorted list of L<namespaces|Gearman::Driver/namespaces>.
+Sends TERM signal to all child processes and exits Gearman::Driver.
 
-=head2 get_modules
+=cut
 
-Returns a sorted list of L<modules|Gearman::Driver/modules>.
+sub shutdown {
+    my ($self) = @_;
+    POE::Kernel->signal( $self->{session}, 'TERM' );
+}
 
-=head2 has_modules
-
-Returns the count of L<modules|Gearman::Driver/modules>.
+sub DEMOLISH {
+    my ($self) = @_;
+    if ( $self->pid eq $$ ) {
+        $self->shutdown;
+        unlink $self->cc_socket;
+    }
+}
 
 =head2 has_job
 
@@ -459,11 +678,6 @@ Returns the job instance.
 sub BUILD {
     my ($self) = @_;
     $self->_setup_logger;
-    $self->_load_namespaces;
-    $self->_start_observer;
-    $self->_start_session;
-
-    # $self->_start_jobs;
 }
 
 sub _setup_logger {
@@ -478,79 +692,52 @@ sub _setup_logger {
     );
 }
 
-sub _load_namespaces {
-    my ($self) = @_;
-
-    my @modules = ();
-    foreach my $ns ( $self->get_namespaces ) {
-        push @modules, useall $ns;
-    }
-
-    unless (@modules) {
-        my $ns = join ', ', $self->get_namespaces;
-        croak "Could not find any modules in those namespaces: $ns";
-    }
-
-    foreach my $module (@modules) {
-        next unless $self->_is_valid_worker_subclass($module);
-        next unless $self->_has_job_method($module);
-        $self->_add_module($module);
-    }
-
-    unless ( $self->has_modules ) {
-        my $modules = join ', ', @modules;
-        croak "None of the modules have a method with 'Job' attribute set: $modules";
-    }
-}
-
-sub _is_valid_worker_subclass {
-    my ( $self, $module ) = @_;
-    return 0 unless $module->can('meta');
-    return 0 unless grep $_ eq 'Gearman::Driver::Worker', $module->meta->superclasses;
-    return 1;
-}
-
-sub _has_job_method {
-    my ( $self, $module ) = @_;
-    return 0 unless $module->meta->can('get_nearest_methods_with_attributes');
-    foreach my $method ( $module->meta->get_nearest_methods_with_attributes ) {
-        next unless grep $_ eq 'Job', @{ $method->attributes };
-        return 1;
-    }
-    return 0;
-}
-
 sub _start_observer {
     my ($self) = @_;
-    $self->{observer} = Gearman::Driver::Observer->new(
-        callback => sub {
-            my ($status) = @_;
-            $self->_observer_callback($status);
-        },
-        interval => $self->interval,
-        server   => $self->server,
-    );
+    if ( $self->interval > 0 ) {
+        $self->{observer} = Gearman::Driver::Observer->new(
+            callback => sub {
+                my ($status) = @_;
+                $self->_observer_callback($status);
+            },
+            interval => $self->interval,
+            server   => $self->server,
+        );
+    }
+}
+
+sub _start_console {
+    my ($self) = @_;
+    if ( $self->console_port > 0 ) {
+        $self->{console} = Gearman::Driver::Console->new(
+            driver => $self,
+            port   => $self->console_port,
+        );
+    }
 }
 
 sub _observer_callback {
     my ( $self, $status ) = @_;
     foreach my $row (@$status) {
         if ( my $job = $self->get_job( $row->{name} ) ) {
-            if ( $job->count_childs && $job->count_childs == $row->{busy} && $row->{queue} ) {
+            if ( $job->count_processes <= $row->{busy} && $row->{queue} ) {
                 my $diff = $row->{queue} - $row->{busy};
-                my $free = $job->max_childs - $job->count_childs;
+                my $free = $job->max_processes - $job->count_processes;
                 if ($free) {
                     my $start = $diff > $free ? $free : $diff;
-                    $job->add_child for 1 .. $start;
+                    $self->log->debug( sprintf "Starting %d new process(es) of type %s", $start, $job->name );
+                    $job->add_process for 1 .. $start;
                 }
             }
-            elsif ( $job->count_childs && $job->count_childs > $job->min_childs && $row->{queue} == 0 ) {
-                my $stop = $job->count_childs - $job->min_childs;
-                $job->remove_child for 1 .. $stop;
-            }
-            elsif ( $job->count_childs < $job->min_childs ) {
-                my $start = $job->min_childs - $job->count_childs;
-                $job->add_child for 1 .. $start;
+
+            elsif ( $job->count_processes && $job->count_processes > $job->min_processes && $row->{queue} == 0 ) {
+                my $idle = time - $self->console->get_lastrun( $job->name );
+                if ( $idle >= $self->max_idle_time ) {
+                    my $stop = $job->count_processes - $job->min_processes;
+                    $self->log->debug( sprintf "Stopping %d process(es) of type %s (idle: %d)",
+                        $stop, $job->name, $idle );
+                    $job->remove_process for 1 .. $stop;
+                }
             }
         }
         else {
@@ -561,11 +748,12 @@ sub _observer_callback {
 
 sub _start_session {
     my ($self) = @_;
-    POE::Session->create(
+    $self->{session} = POE::Session->create(
         object_states => [
             $self => {
-                _start  => '_start',
-                got_sig => '_on_sig',
+                _start            => '_start',
+                got_sig           => '_on_sig',
+                monitor_processes => '_monitor_processes',
             }
         ]
     );
@@ -575,9 +763,9 @@ sub _on_sig {
     my ( $self, $kernel, $heap ) = @_[ OBJECT, KERNEL, HEAP ];
 
     foreach my $job ( $self->get_jobs ) {
-        foreach my $child ( $job->get_childs ) {
-            $self->log->info( sprintf '(%d) [%s] Child killed', $child->PID, $job->name );
-            $child->kill();
+        foreach my $process ( $job->get_processes ) {
+            $self->log->info( sprintf '(%d) [%s] Process killed', $process->PID, $job->name );
+            $process->kill();
         }
     }
 
@@ -588,38 +776,92 @@ sub _on_sig {
 
 sub _start {
     $_[KERNEL]->sig( $_ => 'got_sig' ) for qw(INT QUIT ABRT KILL TERM);
+    $_[OBJECT]->_add_jobs;
     $_[OBJECT]->_start_jobs;
+    $_[KERNEL]->delay( monitor_processes => 5 );
+}
+
+sub _add_jobs {
+    my ($self) = @_;
+
+    foreach my $module ( $self->get_modules ) {
+        my $worker = $module->new( server => $self->server );
+
+        foreach my $method ( $module->meta->get_nearest_methods_with_attributes ) {
+            apply_all_roles( $method => 'Gearman::Driver::Worker::AttributeParser' );
+
+            $method->default_attributes( $worker->default_attributes );
+            $method->override_attributes( $worker->override_attributes );
+
+            next unless $method->has_attribute('Job');
+
+            $self->add_job(
+                {
+                    decode        => $method->get_attribute('Decode'),
+                    encode        => $method->get_attribute('Encode'),
+                    max_processes => $method->get_attribute('MaxProcesses'),
+                    method        => $method->body,
+                    min_processes => $method->get_attribute('MinProcesses'),
+                    name          => $worker->prefix . $method->name,
+                    object        => $worker,
+                }
+            );
+        }
+
+    }
 }
 
 sub _start_jobs {
     my ($self) = @_;
 
-    foreach my $module ( $self->get_modules ) {
-        my $worker = $module->new();
-        foreach my $method ( $module->meta->get_nearest_methods_with_attributes ) {
-            my $attr  = $worker->_parse_attributes( $method->attributes );
-            my $name  = $worker->prefix . $method->name;
-            my $job = Gearman::Driver::Job->new(
-                driver     => $self,
-                method     => $method,
-                name       => $name,
-                worker     => $worker,
-                server     => $self->server,
-                min_childs => $attr->{MinChilds},
-                max_childs => $attr->{MaxChilds},
-            );
-            for ( 1 .. $attr->{MinChilds} ) {
-                $job->add_child();
-            }
-            $self->_set_job( $name => $job );
+    foreach my $job ( $self->get_jobs ) {
+        for ( 1 .. $job->min_processes ) {
+            $job->add_process();
         }
     }
 }
 
+sub _monitor_processes {
+    my $self = $_[OBJECT];
+    foreach my $job ( $self->get_jobs ) {
+        if ( $job->count_processes < $job->min_processes ) {
+            my $start = $job->min_processes - $job->count_processes;
+            $self->log->debug( sprintf "Starting %d new process(es) of type %s", $start, $job->name );
+            $job->add_process for 1 .. $start;
+        }
+    }
+    $_[KERNEL]->delay( monitor_processes => 5 );
+}
+
+no Moose;
+
+__PACKAGE__->meta->make_immutable;
+
+=head1 SCRIPT
+
+There's also a script C<gearman_driver.pl> which is installed with
+this distribution. It just instantiates L<Gearman::Driver> with its
+default values, having most of the options exposed to the command
+line using L<MooseX::Getopt>.
+
+    usage: gearman_driver.pl [long options...]
+            --loglevel          Log level (default: INFO)
+            --lib               Example: --lib ./lib --lib /custom/lib
+            --server            Gearman host[:port][,host[:port]]
+            --logfile           Path to logfile (default: gearman_driver.log)
+            --console_port      Port of management console (default: 47300)
+            --interval          Interval in seconds (see Gearman::Driver::Observer)
+            --loglayout         Log message layout (default: [%d] %p %m%n)
+            --namespaces        Example: --namespaces My::Workers --namespaces My::OtherWorkers
+
 =head1 AUTHOR
 
 Johannes Plunien E<lt>plu@cpan.orgE<gt>
 
+=head1 CONTRIBUTORS
+
+Uwe Voelker, <uwe.voelker@gmx.de>
+
 =head1 COPYRIGHT AND LICENSE
 
 Copyright 2009 by Johannes Plunien
@@ -631,14 +873,28 @@ it under the same terms as Perl itself.
 
 =over 4
 
-=item * L<Gearman::Driver::Observer>
+=item * L<Gearman::Driver::Adaptor>
+
+=item * L<Gearman::Driver::Console>
+
+=item * L<Gearman::Driver::Console::Basic>
+
+=item * L<Gearman::Driver::Console::Client>
 
 =item * L<Gearman::Driver::Job>
 
+=item * L<Gearman::Driver::Loader>
+
+=item * L<Gearman::Driver::Observer>
+
 =item * L<Gearman::Driver::Worker>
 
 =item * L<Gearman::XS>
 
+=item * L<Gearman>
+
+=item * L<Gearman::Server>
+
 =item * L<Log::Log4perl>
 
 =item * L<Module::Find>
